#include <fcntl.h>
#include <linux/soundcard.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <unistd.h>

static int mixer_getinfo( int fd );
static int mixer_getcaps( int fd );

int 
main( void )
{
	int fd;

	if ( ( fd = open( "/dev/mixer", O_RDONLY ) ) == -1 ) {
		perror( "open()" );
		return 1;
	}

	/* ͥ */
	if ( mixer_getinfo( fd ) == -1 ) {
		perror( "mixer_getinfo() failed" );
		close( fd );
		return 1;
	}

	/* ߥǽ(ޤǤ) */
	if ( mixer_getcaps( fd ) == -1 ) {
		perror( "mixer_getcaps() failed" );
		close( fd );
		return 1;
	}

	close( fd );
	return 0;
}
    

/*
 * /dev/mixer ͥ
 */
static int 
mixer_getinfo( int fd )
{
	int devmask, recmask;
	int recsrc;
	int stereodevs;
	int level;
	int i;

	/* soundcard.h Ƥͥ̾ */
	const char* sound_device_names[] = SOUND_DEVICE_NAMES;


	/* ߥݡȤƤͥ */
	if ( ioctl( fd, SOUND_MIXER_READ_DEVMASK, &devmask ) == -1 ) {
		perror( "ioctl( SOUND_MIXER_READ_DEVMASK )" );
		return -1;
	}

	/* ϿȤƻȤͥ */
	if ( ioctl( fd, SOUND_MIXER_READ_RECMASK, &recmask ) == -1 ) {
		perror( "ioctl( SOUND_MIXER_READ_RECMASK )" );
		return -1;
	}

	/* ϿȤ򤵤Ƥͥ */
	if ( ioctl( fd, SOUND_MIXER_READ_RECSRC, &recsrc ) == -1 ) {
		perror( "ioctl( SOUND_MIXER_READ_RECSRC )" );
		return -1;
	}

	/* ͥ뤬ƥ쥪򥵥ݡȤƤ뤫ɤ */
	if ( ioctl( fd, SOUND_MIXER_READ_STEREODEVS, &stereodevs ) == -1 ) {
		perror( "ioctl( SOUND_MIXER_READ_STEREODEVS )" );
		return -1;
	}

	printf("\n" );
	printf("Channel  Support Channel Rec    Current Stereo\n");  
	printf("Name             Level   Source Input         \n");
	printf("-------- ------- ------- ------ ------- ------\n");
            

	/* Ƥͥο롼 */
	for ( i = 0; i < SOUND_MIXER_NRDEVICES; i ++ ) {
		/* ͥ̾ */
		printf( "%-8s ", sound_device_names[i] );

		/* ͥΥݡȾ */
		( devmask & ( 1 << i ) ) ? 
			printf( "      " ) : printf( "      " );


		if ( ( devmask & ( 1 << i ) ) ) {
			/* ͥΥܥ塼 */
			if ( ioctl( fd, MIXER_READ( i ), &level ) == -1 ) {
				perror( "ioctl(MIXER_READ())" );
				return -1; 
			}

			if ( stereodevs & ( 1 << i ) ) {
				/* ƥ쥪ξϲ8ӥåȤ8ӥåȤ */
				printf( "%d%%,%d%% ", 
					level & 0xff, ( level & 0xff00 ) >> 8 );
			}
			else {
				/* Υ */
				printf( "%d%%     ", level & 0xff );
			}
		}
		else {
			printf( "        " );
		}


		/* ϿȤƻȤ뤫ߤϿƥ쥪  */
		( recmask & ( 1 << i ) )    ? 
			printf( "     " )  : printf( "     " );
		
		( recsrc  & ( 1 << i ) )    ? 
			printf( "      " ) : printf( "      " );

		( stereodevs & ( 1 << i ) ) ? 
			printf( "" )   : printf( "" );

		printf( "\n" );
	}

	return 0;
}


/*
 * ɥɤΥߥǽˤĤƤξ
 * ܥץǤϡϿʣƱѲݾΤߡ
 */
static int 
mixer_getcaps( int fd )
{
	int caps;

	/* ɥɤΥߥǽ */
	if ( ioctl( fd, SOUND_MIXER_READ_CAPS, &caps ) == -1 ) {
		perror( "SOUND_MIXER_READ_CAPS" );
		return -1;
	}

	printf( "\n" );


	/* ϿʣƱ˻Ѳǽ */
	( caps & SOUND_CAP_EXCL_INPUT ) ? 
		printf( "### Rec Sources are EXCLUSIVE     ###\n" ) :
		printf( "### Rec Sources are NOT EXCLUSIVE ###\n" );

	printf( "\n" );

	return 0;
}

/* End of mixer_test.c */
