#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>            /* gethostbyname() ... */
#include <netinet/in.h>       /* htons() ... */
#include <arpa/inet.h>        /* inet_aton() ... */
#include <linux/soundcard.h>


static int setup_dsp( int* dsp );

static int setup_udp( 
	int* sock, struct sockaddr_in* server_addr, 
	char* host, unsigned short port );

static int need_to_skip( unsigned char* buf, int len );


/*
 * Ф UDP ݡ
 */
#define PORT     13000


/*
 * ̻Ҳӥåȿ : 16 bits
 * ץ󥰼ȿ : 44.1 KHz
 * ͥ : 1
 *
 * ե饰ȥ 100 ms ʬΰ
 *
 * 16 (bits) * 44100 (Hz) * 0.1 (sec) = 8820 (bytes)
 *
 * ۤʤ 2 ζҾȤ 
 * ( SNDCTL_DSP_SETFRAGMENT λͤˤ )
 */
#define FRAGSIZE 8192


int 
main( int argc, char** argv )
{
	struct sockaddr_in server_addr;
	int sock;
	int dsp;
	int len;

	unsigned char buf[ FRAGSIZE ];

	if ( argc != 2 ) {
		fprintf( stderr, "usage : client host\n" );
		return 1;
	}


	/* /dev/dsp  */
	if ( setup_dsp( &dsp ) != 0 ) {
		fprintf( stderr, "Setup /dev/dsp failed.\n" );
		return 1;
	}

	/* UDP åȤ */
	if ( setup_udp( &sock, &server_addr, argv[1], PORT ) != 0 ) {
		fprintf( stderr, "Setup UDP socket failed.\n" );
		close( dsp );
		return 1;
	}


	/*
	 * 顼뤫λޤǥ롼פޤ
	 */
	while ( 1 ) {
		/*
		 * FRAGSIZE Ϥɤ߹ޤǥ֥å
		 */
		if ( ( len = read( dsp, buf, sizeof( buf ) ) ) == -1 ) {
			perror( "read()" );
			close( dsp  );
			close( sock );
			return 1;
		}

		if ( ! need_to_skip( buf, len ) ) {
			/*
			 * ɤ߹Ϥ UDP ǥФ
			 */
			if ( sendto( sock, buf, len, 0, 
				     ( struct sockaddr* )&server_addr, 
				     sizeof( struct sockaddr_in ) ) == -1 ) {

				perror( "sendto()" );
				close( dsp  );
				close( sock );
				return 1;
			}
		}
	}

	/*
	 * while() 롼פϽʤΤǤˤޤ󡣡
	 */
	close( dsp  );
	close( sock );

	return 0;
}


/*
 * /dev/dsp ꤹ
 */
static 
int setup_dsp( int* dsp )
{
	int fmt     = AFMT_S16_LE;
	int freq    = 44100;
	int channel = 1;

	int bit;
	int i;
	int fragarg;


	if ( ( *dsp = open( "/dev/dsp", O_RDONLY ) ) == -1 ) {
		perror( "open()" );
		return -1;
	}

	/* 
	 * ե饰ȥ
	 *
	 * 0xMMMMSSSS ( MMMM ϥե饰ȿSSSS 
	 * ե饰ȥ = 2^{SSSS} Ȥʤ )
	 *
	 * ηꤹ뤿ᡢޤ SSSS 롣
	 *
	 * ( 0 <= SSSS <= 65535, ե饰ȥ
	 *   ϺǤХåե / 2 ޤǤ
	 *   Ǥʤ )
	 *
	 * ޤ MMMM ä¤ߤʤ 0x7fff
	 * ꤹ롣
	 */
	for ( i = 0, bit = 1; i <= 65535; bit <<= 1, i ++ ) {
		if ( FRAGSIZE < bit ) {
			i = ( i == 0 ) ? 0 : i - 1;
			break;
		}
	}

	fragarg = ( 0x7fff << 16 ) | i;

	if ( ioctl( *dsp, SNDCTL_DSP_SETFRAGMENT, &fragarg )== -1 ) {
		perror( "ioctl( SNDCTL_DSP_SETFRAGMENT )" );
		close( *dsp );
		return -1;
	}

	if ( ioctl( *dsp, SOUND_PCM_SETFMT, &fmt ) == -1 ) {
		perror( "ioctl( SOUND_PCM_SETFMT )" );
		close( *dsp );
		return -1;
	}

	if ( ioctl( *dsp, SOUND_PCM_WRITE_CHANNELS, &channel ) == -1 ) {
		perror( "ioctl( SOUND_PCM_WRITE_CHANNELS )" );
		close( *dsp );
		return -1;
	}

	if ( ioctl( *dsp, SOUND_PCM_WRITE_RATE, &freq ) == -1 ) {
		perror( "ioctl( SOUND_PCM_WRITE_RATE )" );
		close( *dsp );
		return -1;
	}
          
	return 0;
}


/*
 * ФȤ̿Ѥ UDP åȤꤹ
 */
static int 
setup_udp( 
	int* sock, struct sockaddr_in* server_addr, 
	char* host, unsigned short port )
{
	struct in_addr     inaddr;
	struct hostent*    ent;

	if ( ( *sock = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 ) {
		perror( "socket()" );
		return -1;
	}

	/*
	 * host  IP ɥ쥹ǤȲꤷ inet_aton() ˤ
	 * ʸɽ IP ɥ쥹ХʥͤѴ롣
	 *
	 * Ԥϥۥ̾ǤȲꤷ
	 * gethostbyname() ˤäбۥȾ
	 * ( ξ硢/etc/hosts ե뤫 DNS ˤä̾
	 *   Ǥɬפ )
	 */
	if ( ! inet_aton( host, &inaddr ) ) {
		ent = gethostbyname( host );

		if ( ent == NULL ) {
			fprintf( stderr, 
				 "Couldn't get host information specified by %s",
				 host );

			close( *sock );			
			return -1;
		}

		inaddr = *( struct in_addr* )( ent->h_addr_list[0] );
	}

	memset( server_addr, 0, sizeof( struct sockaddr_in ) );

	server_addr->sin_family = AF_INET;
	server_addr->sin_port   = htons( port );

	memcpy( &server_addr->sin_addr, &inaddr.s_addr, 
		sizeof( server_addr->sin_addr ) );

	return 0;
}


/*
 * ϥХåեĴ٤ơ̵֤˶ᤤɤȽꤹ롣
 */
static int 
need_to_skip( unsigned char* buf, int len )
{
	/*
	 * ͰʾΥǡ¸ߤʤϥФʤ
	 * ʬδ塢̵٤ŬͤǤ
	 */
	int threshold = 0x0040;

	int i;
	unsigned short value;

	for ( i = 0; i < len; i += 2 ) {
		memcpy( &value, &buf[i], sizeof( value ) );
		if ( value >= threshold )
			return 0;
	}

	return 1;
}


/* End of client.c */
