#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>            /* gethostbyname() ... */
#include <netinet/in.h>       /* htons() ... */
#include <arpa/inet.h>        /* inet_aton() ... */
#include <linux/soundcard.h>


static int setup_dsp( int* dsp  );
static int setup_udp( int* sock, unsigned short port );


/*
 * Ф UDP ݡ
 */
#define PORT     13000


/*
 * ̻Ҳӥåȿ : 16 bits
 * ץ󥰼ȿ : 44.1 KHz
 * ͥ : 1
 *
 * ե饰ȥ 100 ms ʬΰ
 *
 * 16 (bits) * 44100 (Hz) * 0.1 (sec) = 8820 (bytes)
 *
 * ۤʤ 2 ζҾȤ 
 * ( SNDCTL_DSP_SETFRAGMENT λͤˤ )
 */
#define FRAGSIZE 8192


int 
main( int argc, char** argv )
{
	struct sockaddr_in client_addr;
	int sock;
	int dsp;
	int len;
	int addr_len;

	unsigned char buf[ FRAGSIZE ];


	/* /dev/dsp  */
	if ( setup_dsp( &dsp ) != 0 ) {
		fprintf( stderr, "Setup /dev/dsp failed.\n" );
		return 1;
	}

	/* UDP åȤ */
	if ( setup_udp( &sock, PORT ) != 0 ) {
		fprintf( stderr, "Setup UDP socket failed.\n" );
		close( dsp );
		return 1;
	}

	addr_len = sizeof( struct sockaddr_in );

	/*
	 * 顼뤫λޤǥ롼פޤ
	 */
	while ( 1 ) {
		/*
		 * UDP åϤɤ߹ޤǥ֥å
		 */
		if ( ( len = recvfrom( sock, buf, sizeof( buf ), 0, 
		       ( struct sockaddr* )&client_addr, &addr_len ) ) == -1 ) {

			perror( "recvfrom()" );
			close( dsp  );
			close( sock );
			return 1;
		}

		/*
		 * /dev/dsp ˽񤭹
		 */
		if ( write( dsp, buf, len ) == -1 ) {
			perror( "write()" );
			close( dsp  );
			close( sock );
			return 1;
		}
	}


	/*
	 * while() 롼פϽʤΤǤˤޤ󡣡
	 */
	close( dsp  );
	close( sock );

	return 0;
}


/*
 * /dev/dsp ꤹ
 */
static int 
setup_dsp( int* dsp )
{
	int fmt     = AFMT_S16_LE;
	int freq    = 44100;
	int channel = 1;

	int bit;
	int i;
	int fragarg;


	if ( ( *dsp = open( "/dev/dsp", O_WRONLY ) ) == -1 ) {
		perror( "open()" );
		return -1;
	}

	/* 
	 * ե饰ȥ
	 *
	 * 0xMMMMSSSS ( MMMM ϥե饰ȿSSSS 
	 * ե饰ȥ = 2^{SSSS} Ȥʤ )
	 *
	 * ηꤹ뤿ᡢޤ SSSS 롣
	 *
	 * ( 0 <= SSSS <= 65535, ե饰ȥ
	 *   ϺǤХåե / 2 ޤǤ
	 *   Ǥʤ )
	 *
	 * ޤ MMMM ä¤ߤʤ 0x7fff
	 * ꤹ롣
	 */
	for ( i = 0, bit = 1; i <= 65535; bit <<= 1, i ++ ) {
		if ( FRAGSIZE < bit ) {
			i = ( i == 0 ) ? 0 : i - 1;
			break;
		}
	}

	fragarg = ( 0x7fff << 16 ) | i;

	if ( ioctl( *dsp, SNDCTL_DSP_SETFRAGMENT, &fragarg )== -1 ) {
		perror( "ioctl( SNDCTL_DSP_SETFRAGMENT )" );
		close( *dsp );
		return -1;
	}

	if ( ioctl( *dsp, SOUND_PCM_SETFMT, &fmt ) == -1 ) {
		perror( "ioctl( SOUND_PCM_SETFMT )" );
		close( *dsp );
		return -1;
	}

	if ( ioctl( *dsp, SOUND_PCM_WRITE_CHANNELS, &channel ) == -1 ) {
		perror( "ioctl( SOUND_PCM_WRITE_CHANNELS )" );
		close( *dsp );
		return -1;
	}

	if ( ioctl( *dsp, SOUND_PCM_WRITE_RATE, &freq ) == -1 ) {
		perror( "ioctl( SOUND_PCM_WRITE_RATE )" );
		close( *dsp );
		return -1;
	}
          
	return 0;
}


/*
 * 饤ȤȤ̿Ѥ UDP åȤꤹ
 */
static int 
setup_udp( int* sock, unsigned short port )
{
	struct sockaddr_in server_addr;

	if ( ( *sock = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 ) {
		perror( "socket()" );
		return -1;
	}

	memset( &server_addr, 0, sizeof( struct sockaddr_in ) );

	server_addr.sin_family      = AF_INET;
	server_addr.sin_addr.s_addr = INADDR_ANY;
	server_addr.sin_port        = htons( port );

	if ( bind( *sock, ( struct sockaddr* )&server_addr, 
		   sizeof( struct sockaddr_in ) ) == -1 ) {

		perror( "bind()" );
		close( *sock );
		return -1;
	}

	return 0;
}


/* End of server.c */
