#include <fcntl.h> 
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <linux/soundcard.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <unistd.h>


#define COMMAND_STOP 0
#define COMMAND_PLAY 1
#define COMMAND_EXIT 2

#define PAN_MIN  0
#define PAN_MAX  256
#define PAN_STEP 4

#define BUFSIZE  4096


/* WAVE եξǼ빽¤ */
typedef struct {
	FILE* fp;        /* ե빽¤ */
	short is_pcm;    /* PCM եޥåȤξ 1ʳ 0 */
	short channel;   /* Υξ 1ƥ쥪ξ 2 */
	int   rate;      /* ץ󥰼ȿ */
	short bits;      /* ̻Ҳӥåȿ */
	long  offset;    /* եƬ PCM ǡޤǤΥեå */
	int   len;       /* PCM ǡĹ */
	int   current;   /* ߤκ */
}
WAVE;


/* ɬפʥǡޤȤ᤿¤ */
typedef struct {
	pthread_t thread_id;
	int command;
	int dsp;
	WAVE wave;
	short buffer[ BUFSIZE / sizeof( short ) ];
	/* 
	 * PAN_MIN(0) <= pan <= PAN_MAX(256)
	 * pan == PAN_MIN  L Τߤ pan == PAN_MAX  R Τ
	 */
	int pan;
}
Context;

static Context context;


static int   read_file( char* fname, WAVE* wave );
static int   setup_dsp( int* fd, WAVE* wave );
static void  show_menu( void );
static void* play_thread( void* param );
static void  fill_buffer( short* dst, WAVE* wave, int pan );
static void  shutdown();


int 
main( int argc, char** argv )
{
	pthread_attr_t attr;
	int ret;

	if ( argc != 2 ) {
		fprintf( stderr, "usage : ./wave filename\n" );
		return 1;
	}

	if ( read_file( argv[1], &( context.wave ) ) != 0 ) {
		fprintf( stderr, "Failed to read wave files.\n" );
		return 1;
	}

	if ( setup_dsp( &( context.dsp ), &( context.wave ) ) != 0 ) {
		fprintf( stderr, "Failed to setup /dev/dsp.\n" );
		return 1;
	}

	pthread_attr_init( &attr );
	pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_JOINABLE );

	ret = pthread_create( &( context.thread_id ), &attr, 
			      play_thread, ( void* )( &context ) );

	pthread_attr_destroy( &attr );

	if ( ret ) {
		fprintf( stderr, "Failed to create thread.\n" );
		shutdown();
		return 1;
	}

	show_menu();
	pthread_join( context.thread_id, NULL );
	shutdown();

	return 0;
}


static int 
read_file( char* fname, WAVE* wave )
{
	char buf[32];
	int  len;

	if ( ( wave->fp = fopen( fname, "r" ) ) == NULL ) {
		fprintf( stderr, "Failed to open %s\n", fname );
		goto err;
	}

	/* 
	 * Ƭ 4 ХȤ "RIFF" Ǥ뤳Ȥǧ
	 *  4 ХȥåפƤ
	 */
	fread( buf, 8, 1, wave->fp );

	if ( strncmp( buf, "RIFF", 4 ) != 0 ) {
		fprintf( stderr, "Specified file is not RIFF file.\n" );
		goto err;
	}


	/*  4 ХȤ "WAVE" Ǥ뤳Ȥǧ */
	fread( buf, 4, 1, wave->fp );

	if ( strncmp( buf, "WAVE", 4 ) != 0 ) {
		fprintf( stderr, "Specified file is not WAVE file.\n" );
		goto err;
	}

	/* fmt 󥯤õ */
	while ( 1 ) {
		fread( buf, 8, 1, wave->fp );
		len = *( int* )( &buf[4] );

		if ( strncmp( buf, "fmt ", 4 ) != 0 ) {
			if ( fseek( wave->fp, len, SEEK_CUR ) == -1 ) {
				fprintf( stderr, "Failed to find fmt chunk.\n" );
				goto err;
			}
		}
		else {
			break;
		}
	}

	fread( buf, len, 1, wave->fp );

	wave->is_pcm  = *( ( short* )( &buf[0]  ) );
	wave->channel = *( ( short* )( &buf[2]  ) );
	wave->rate    = *( ( int*   )( &buf[4]  ) );
	wave->bits    = *( ( short* )( &buf[14] ) );

	/*
	 * PCM / 16 bits / Monoral ʳ WAVE ǡˤĤƤ
	 * ܥץǤоݳ
	 */
	if ( wave->is_pcm != 1 ) {
		fprintf( stderr, "WAVE file must be PCM format.\n" );
		goto err;
	}

	if ( wave->bits != 16 ) {
		fprintf( stderr, "WAVE file must be 16 bits .\n" );
		goto err;
	}

	if ( wave->channel != 1 ) {
		fprintf( stderr, "WAVE file must be monoral.\n" );
		goto err;
	}

	/* data 󥯤õ */
	while ( 1 ) {
		fread( buf, 8, 1, wave->fp );
		len = *( int* )( &buf[4] );

		if ( strncmp( buf, "data", 4 ) != 0 ) {
			if ( fseek( wave->fp, len, SEEK_CUR ) == -1 ) {
				fprintf( stderr, "Failed to find data chunk.\n" );
				goto err;
			}
		}
		else {
			break;
		}
	}

	wave->len = len;

	if ( ( wave->offset = ftell( wave->fp ) ) == -1 ) {
		fprintf( stderr, "Failed to find offset of PCM data.\n" );
		goto err;
	}

	printf( "\n" );
	printf( "WAVE file format of %s:\n", fname );
	printf( " rate    = %d\n",  wave->rate );
	printf( " channel = %d\n",  wave->channel );
	printf( " bits    = %d\n",  wave->bits );
	printf( " offsets = %ld\n", wave->offset );
	printf( " length  = %d\n",  wave->len );
	printf( "\n" );

	return 0;

err:
	if ( wave->fp != NULL ) {
		fclose( wave->fp );
	}

	return 1;
}


static int 
setup_dsp( int* dsp, WAVE* wave )
{
	int fragment = 0x0004000c; /* size = 4096 bytes, num of fragments = 4. */
	int format = AFMT_S16_LE;
	int rate;
	int channel = 2;

	rate = ( int )wave->rate;

	if ( ( *dsp = open( "/dev/dsp", O_WRONLY ) ) == -1 ) {
		perror( "open()" );
		return 1;
	}

	if ( ioctl( *dsp, SNDCTL_DSP_SETFRAGMENT, &fragment ) == -1 ) {
		perror( "ioctl( SNDCTL_DSP_SETFRAGMENT )" );
		close( *dsp );
		return 1;
	}

	if ( ioctl( *dsp, SNDCTL_DSP_SETFMT, &format ) == -1 ) {
		perror( "ioctl( SNDCTL_DSP_SETFMT )" );
		close( *dsp );
		return 1;
	}

	if ( ioctl( *dsp, SNDCTL_DSP_SPEED, &rate ) == -1 ) {
		perror( "ioctl( SNDCTL_DSP_SPEED )" );
		close( *dsp );
		return 1;
	}

	if ( ioctl( *dsp, SNDCTL_DSP_CHANNELS, &channel ) == -1 ) {
		perror( "ioctl( SNDCTL_DSP_CHANNELS )" );
		close( *dsp );
		return 1;
	}

	return 0;
}


static void 
show_menu( void )
{
	char line[4];
	int  command;

	while ( 1 ) {
		printf( "Select menu ...\n" );
		printf( "  1. Play specified wave file.\n" );
		printf( "  2. Stop playing.\n" );
		printf( " 99. Exit.\n" );
		printf( "\n" );
		printf( "Command ? " );

		fgets( line, sizeof( line ), stdin );
		sscanf( line, "%d\n", &command );

		switch ( command ) {
			case 1:
				context.command = COMMAND_PLAY;
				break;
			case 2:
				context.command = COMMAND_STOP;
				break;
			case 99:
				context.command = COMMAND_EXIT;
				return;
		}
	}
}


static void*
play_thread( void* param )
{
	static int sign = 1;
	Context* context = ( Context* )param;

	while ( context->command != COMMAND_EXIT ) {
		memset( ( void* )( context->buffer ), 0, BUFSIZE );

		switch ( context->command ) {
			case COMMAND_PLAY:
				context->pan += PAN_STEP * sign;

				if ( context->pan > PAN_MAX ) {
					context->pan = PAN_MAX;
					sign *= -1;
				}
				else if ( context->pan < 0 ) {
					context->pan = 0;
					sign *= -1;
				}

				fill_buffer( ( void* )( context->buffer ), 
					     &( context->wave ), context->pan );
				break;

			case COMMAND_STOP:
			default:
				break;
		}

		write( context->dsp, ( void* )( context->buffer ), BUFSIZE );
	}

	pthread_exit( NULL );
}


static void
fill_buffer( short* dst, WAVE* wave, int pan )
{
	/* 
	 * buf ϲƱƤž뤿ΥХåե
	 * BUFSIZE ϽѥХåեǤꡢϤϥƥ쥪/ϤϥΥ
	 * Ǥ뤿ᡢBUFSIZE / 2 ȤƤ롣
	 */
	short buf[ ( BUFSIZE / 2 ) / sizeof( short ) ];
	int copied = 0;
	int i;

	while ( 1 ) {
		int src_left = wave->len - wave->current;
		int dst_left = BUFSIZE / 2 - copied;
		int copy_len = src_left > dst_left ? dst_left : src_left;

		if ( src_left == 0 ) {
			fseek( wave->fp, wave->offset, SEEK_SET );
			wave->current = 0;
			continue;
		}

		if ( dst_left == 0 ) {
			break;
		}

		fread( ( void* )( buf + copied / sizeof( short ) ), 
		       copy_len, 1, wave->fp );

		for ( i = 0; i < copy_len / sizeof( short ); i ++ ) {
			dst[ ( copied / sizeof( short ) + i ) * 2 + 0 ] = 
				buf[i] * ( PAN_MAX - pan ) / PAN_MAX; /* L */
			dst[ ( copied / sizeof( short ) + i ) * 2 + 1 ] =
				buf[i] * pan / PAN_MAX;               /* R */
		}

		wave->current += copy_len;
		copied += copy_len;
	}
}


static void
shutdown( void )
{
	fclose( context.wave.fp );
	close( context.dsp );
}


/* End of pan.c */
